# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model transformation logic for the 'rename id property' transformation"""

from schemapack.spec.schemapack import IDSpec, SchemaPack

from metldata.builtin_transformations.common.utils import model_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def rename_model_id_property(
    *,
    model: SchemaPack,
    class_name: str,
    id_property_name: str,
    description: str | None,
) -> SchemaPack:
    """Rename the id property of the model."""
    mutable_model = model_to_dict(model)
    try:
        mutable_model["classes"][class_name]["id"] = IDSpec(
            propertyName=id_property_name,
            description=description,
        )
    except KeyError as exc:
        raise EvitableTransformationError() from exc
    return SchemaPack.model_validate(mutable_model)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Resolve relation paths (active, passive and complex) and infer target resources of the overall path."

from schemapack.spec.custom_types import ResourceId
from schemapack.spec.datapack import DataPack

from metldata.builtin_transformations.common.path import (
    RelationPath,
)
from metldata.builtin_transformations.common.path.path_elements import (
    RelationPathElement,
    RelationPathElementType,
)
from metldata.transform.exceptions import EvitableTransformationError


def resolve_active_path_element(
    *,
    data: DataPack,
    source_resource_id: ResourceId,
    path_element: RelationPathElement,
) -> frozenset[ResourceId]:
    """Resolve the given relation inference path element of active type for the given
    source resource.

    Args:
        data:
            The data pack to look up resources in.
        source_resource_id:
            The id of the resource to which the path element is applied.
        path_element:
            The relation inference path element to resolve. It is assumed to be of
            active type.

    Returns:
        A set of resource IDs that are targeted by the path element in context of the
        given source resource.
    """
    if path_element.type_ != RelationPathElementType.ACTIVE:
        raise ValueError(
            "Expected path element of type 'ACTIVE', but got a 'PASSIVE' one."
        )

    source_resource = data.resources.get(path_element.source, {}).get(
        source_resource_id
    )

    if not source_resource:
        raise EvitableTransformationError()
    target_resource_relations = source_resource.relations.get(path_element.property)
    target_resource_ids = (
        target_resource_relations.targetResources
        if target_resource_relations
        else frozenset()
    )
    if target_resource_ids is None:
        target_resource_ids = frozenset()
    elif isinstance(target_resource_ids, str):
        target_resource_ids = frozenset({target_resource_ids})
    return target_resource_ids


def resolve_passive_path_element(
    *,
    data: DataPack,
    source_resource_id: ResourceId,
    path_element: RelationPathElement,
) -> frozenset[ResourceId]:
    """Resolve the given relation inference path element of passive type for the given
    source resource.

    Args:
        data:
            The data pack to look up resources in.
        source_resource_id:
            The id of the resource to which the path element is applied.
        path_element:
            The relation inference path element to resolve. It is assumed to be of
            passive type.

    Returns:
        A set of resource IDs that are targeted by the path element in context of the
        given source resource.
    """
    if path_element.type_ != RelationPathElementType.PASSIVE:
        raise ValueError(
            "Expected path element of type 'PASSIVE', but got an 'ACTIVE' one."
        )
    candidate_resources = data.resources.get(path_element.target, {})
    target_resource_ids = set()

    for candidate_resource_id, candidate_resource in candidate_resources.items():
        candidate_resource_relations = candidate_resource.relations.get(
            path_element.property
        )
        relation = (
            candidate_resource_relations.targetResources
            if candidate_resource_relations
            else frozenset()
        )
        if (
            isinstance(relation, frozenset) and source_resource_id in relation
        ) or source_resource_id == relation:
            target_resource_ids.add(candidate_resource_id)

    return frozenset(target_resource_ids)


def resolve_path_element(
    *,
    data: DataPack,
    source_resource_id: ResourceId,
    path_element: RelationPathElement,
) -> frozenset[ResourceId]:
    """Resolve the given relation inference path element for the given source resource.

    Args:
        data: The data pack to look up resources in.
        source_resource_id: The id of the resource to which the path element is applied.
        path_element: The relation inference path element to resolve.

    Returns:
        A set of resource IDs that are targeted by the path element in context of the
        given source resource.
    """
    resolve = (
        resolve_active_path_element
        if path_element.type_ == RelationPathElementType.ACTIVE
        else resolve_passive_path_element
    )
    return resolve(
        data=data,
        source_resource_id=source_resource_id,
        path_element=path_element,
    )


def resolve_path(
    *, data: DataPack, source_resource_id: ResourceId, path: RelationPath
) -> frozenset[ResourceId]:
    """Resolve the given relation inference path for the given source resource.

    Args:
        data: The data pack to look up resources in.
        source_resource_id: The id of the resource to which the path is applied.
        path: The relation inference path to resolve.

    Returns:
        A set of resource IDs that are targeted by the path in context of the given
        source resource.
    """
    resource_ids: set[ResourceId] = {source_resource_id}
    for path_element in path.elements:
        # the target resources of the current iteration are the source resources of the
        # next iteration:
        resource_ids = {
            target_resource_id
            for source_resource_id in resource_ids
            for target_resource_id in resolve_path_element(
                data=data,
                source_resource_id=source_resource_id,
                path_element=path_element,
            )
        }
    return frozenset(resource_ids)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic for transforming data."""

from schemapack.spec.datapack import DataPack

from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def duplicate_data_class(
    *, data: DataPack, source_class_name: str, target_class_name: str
) -> DataPack:
    """Copy the contents of a class in the given DataPack under a new class name.

    Args:
        data:
            The DataPack instance containing the source class to duplicate.
        transformation_config:
            Configuration specifying the source and target class for the copy operation.
    Returns:
        A new DataPack with the source class content duplicated under the target class name.
    """
    modified_data = data_to_dict(data)

    try:
        source_class_resources = modified_data["resources"][source_class_name]
        modified_data["resources"][target_class_name] = source_class_resources
    except KeyError as exc:
        raise EvitableTransformationError() from exc

    return DataPack.model_validate(modified_data)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Models to describe transformations and workflows."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import TypeVar

from pydantic import (
    BaseModel,
    Field,
)
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

Config = TypeVar("Config", bound=BaseModel)
SubmissionAnnotation = TypeVar("SubmissionAnnotation", bound=BaseModel)


class DataTransformer[Config, SubmissionAnnotation](ABC):
    """A base class for a data transformer."""

    def __init__(
        self, *, config: Config, input_model: SchemaPack, transformed_model: SchemaPack
    ):
        """Initialize the transformer with config params, the input model, and the
        transformed model.
        """
        self._config = config
        self._input_model = input_model
        self._transformed_model = transformed_model

    @abstractmethod
    def transform(self, data: DataPack, annotation: SubmissionAnnotation) -> DataPack:
        """Transforms data.

        Args:
            data: The data as DataPack to be transformed.

        Raises:
            DataTransformationError:
                if the transformation fails.
        """
        ...


class TransformationDefinition[Config](BaseModel):
    """A model for describing a transformation."""

    config_cls: type[Config] = Field(
        default=..., description="The config model of the transformation."
    )
    check_model_assumptions: Callable[[SchemaPack, Config], None] = Field(
        default=...,
        description=(
            "A function that checks the assumptions made about the input model."
            + " Raises a ModelAssumptionError if the assumptions are not met."
        ),
    )
    transform_model: Callable[[SchemaPack, Config], SchemaPack] = Field(
        default=...,
        description=(
            "A function to transform the model. Raises a"
            + " ModelTransformationError if the transformation fails."
        ),
    )
    data_transformer_factory: type[DataTransformer] = Field(
        default=...,
        description=(
            "A class for transforming data. Raises a DataTransformationError"
            + " if the transformation fails."
        ),
    )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Logic for handling Transformation."""

from typing import override

import schemapack.exceptions
from schemapack import SchemaPackValidator
from schemapack._internals.validation.base import (
    ClassValidationPlugin,
    GlobalValidationPlugin,
    ResourceValidationPlugin,
)
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.transform.base import (
    Config,
    TransformationDefinition,
)
from metldata.transform.exceptions import (
    PostTransformValidationError,
    PreTransformValidationError,
)


class NoOpValidator(SchemaPackValidator):
    """Custom no-op variant of the SchemaPackValidator used for skipping validation."""

    @override
    def __init__(
        self,
        *,
        schemapack: SchemaPack,
        add_global_plugins: list[type[GlobalValidationPlugin]] | None = None,
        add_class_plugins: list[type[ClassValidationPlugin]] | None = None,
        add_resource_plugins: list[type[ResourceValidationPlugin]] | None = None,
    ):
        """Do nothing with the input, just skip plugin creation."""

    @override
    def validate(self, *, datapack: DataPack):
        """Do nothing and return."""
        return


class TransformationHandler[SubmissionAnnotation]:
    """Used for executing transformations described in a TransformationDefinition."""

    def __init__(
        self,
        transformation_definition: TransformationDefinition[Config],
        transformation_config: Config,
        input_model: SchemaPack,
        validate_input: bool = False,
        validate_output: bool = False,
    ):
        """Initialize the TransformationHandler by checking the assumptions made on the
        input model and transforming the model as described in the transformation
        definition. The transformed model is available at the `transformed_model`
        attribute.

        Raises:
            ModelAssumptionError:
                if the assumptions made on the input model are not met.
        """
        self._definition = transformation_definition
        self._config = transformation_config
        self._input_model = input_model

        self._definition.check_model_assumptions(self._input_model, self._config)
        self.transformed_model = self._definition.transform_model(
            self._input_model, self._config
        )
        self._data_transformer = self._definition.data_transformer_factory(
            config=self._config,
            input_model=self._input_model,
            transformed_model=self.transformed_model,
        )
        self._input_data_validator = (
            SchemaPackValidator(schemapack=self._input_model)
            if validate_input
            else NoOpValidator(schemapack=self._input_model)
        )
        self._transformed_data_validator = (
            SchemaPackValidator(schemapack=self.transformed_model)
            if validate_output
            else NoOpValidator(schemapack=self.transformed_model)
        )

    def transform_data(
        self, data: DataPack, annotation: SubmissionAnnotation
    ) -> DataPack:
        """Transforms data using the transformation definition. Validates the
        input data against the input model and the transformed data
        against the transformed model.

        Args:
            data: The data to be transformed.

        Raises:
            PreTransformValidation:
                If validation of input data fails against the input model.
            PostTransformValidation:
                If validation of transformed data fails against the transformed model.
            DataTransformationError:
                if the transformation fails.
        """
        try:
            self._input_data_validator.validate(datapack=data)
        except schemapack.exceptions.ValidationError as error:
            raise PreTransformValidationError(validation_error=error) from error

        transformed_data = self._data_transformer.transform(
            data=data, annotation=annotation
        )

        try:
            self._transformed_data_validator.validate(datapack=transformed_data)
        except schemapack.exceptions.ValidationError as error:
            raise PostTransformValidationError(validation_error=error) from error

        return transformed_data



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data transformation logic for the 'replace resource ids' transformation"""

from arcticfreeze import FrozenDict
from pydantic import BaseModel
from schemapack.spec.datapack import DataPack, Resource

from metldata.builtin_transformations.common.custom_types import (
    AccessionMap,
    MutableDatapack,
    ResourceId,
)
from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def replace_data_resource_ids(
    *,
    data: DataPack,
    class_name: str,
    annotation: BaseModel,
) -> DataPack:
    """Replace resource ids of a data class using annotations."""
    modified_data = data_to_dict(data)

    resource_accessions = _get_resource_accessions(
        annotation=annotation, class_name=class_name
    )

    original_resources = data.resources.get(class_name)
    if original_resources is None:
        raise EvitableTransformationError()

    updated_resources = _update_resources(
        original_resources=original_resources,
        resource_accessions=resource_accessions,
        class_name=class_name,
    )

    modified_data["resources"][class_name] = updated_resources

    _replace_ids_in_relations(
        modified_data=modified_data,
        original_data=data,
        target_class_name=class_name,
        resource_accessions=resource_accessions,
    )
    return DataPack.model_validate(modified_data)


def _get_resource_accessions(*, class_name: str, annotation: BaseModel) -> AccessionMap:
    """Extract resource ids from annotations."""
    try:
        accession_map = annotation.model_dump()["accession_map"]
    except KeyError as exc:
        raise ValueError(
            "The annotation is missing the required 'accession_map' field. "
            "Expected structure: {'accession_map': {<class_name>: {<old_id>: <new_id>, ...}, ...}}"
        ) from exc

    resource_accessions = accession_map.get(class_name)
    if resource_accessions is None:
        raise ValueError(f"No accession map found for class '{class_name}'.")

    return resource_accessions


def _update_resources(
    *,
    original_resources: FrozenDict[ResourceId, Resource],
    resource_accessions: AccessionMap,
    class_name: str,
) -> dict[ResourceId, Resource]:
    """Update resources with new resource ids."""
    updated_resources = {
        new_id: original_resources[old_id]
        for old_id, new_id in resource_accessions.items()
        if old_id in original_resources
    }
    # Set operation uses the keys
    missing_ids = set(original_resources) - set(resource_accessions)
    if missing_ids:
        raise ValueError(
            f"The following original resource IDs from class '{class_name}' are missing in the annotation's accession map: {missing_ids}."
        )
    return updated_resources


def _replace_ids_in_relations(
    *,
    modified_data: MutableDatapack,
    original_data: DataPack,
    resource_accessions: AccessionMap,
    target_class_name: str,
) -> None:
    """Replace resource IDs in relations of the modified data.

    Args:
        modified_data (dict): Dictionary representation of the modified data.
        original_data (DataPack): Original data.
        target_class (str): Class name whose resource ids are updated.
    """
    for class_name, class_resources in original_data.resources.items():
        for resource_id, resource in class_resources.items():
            relations = resource.relations
            if not relations:
                continue

            for relation_name, relation_spec in resource.relations.items():
                new_target_ids: str | set
                if relation_spec.targetClass != target_class_name:
                    continue

                target_resources = relation_spec.targetResources

                if not target_resources:
                    continue

                if isinstance(target_resources, str):
                    new_target_ids = resource_accessions[target_resources]
                elif isinstance(target_resources, frozenset):
                    new_target_ids = {
                        resource_accessions[target_resource]
                        for target_resource in target_resources
                    }

                modified_data["resources"][class_name][resource_id]["relations"][
                    relation_name
                ] = {
                    "targetClass": target_class_name,
                    "targetResources": new_target_ids,
                }



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"A transformation to merge relations."

from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.merge_relations.assumptions import (
    check_model_assumptions,
)
from metldata.builtin_transformations.merge_relations.config import (
    MergeRelationsConfig,
)
from metldata.builtin_transformations.merge_relations.data_transform import (
    merge_data_relations,
)
from metldata.builtin_transformations.merge_relations.model_transform import (
    RelationSpecificationParams,
    merge_model_relations,
)
from metldata.builtin_transformations.registry import register_transformation
from metldata.transform.base import (
    DataTransformer,
    SubmissionAnnotation,
    TransformationDefinition,
)


class MergeRelationsTransformer(
    DataTransformer[MergeRelationsConfig, SubmissionAnnotation]
):
    """A transformer that merges relations within data."""

    def transform(self, data: DataPack, annotation: SubmissionAnnotation) -> DataPack:
        """Transforms data.

        Args:
            data: The data as DataPack to be transformed.

        Raises:
            DataTransformationError:
                if the transformation fails.
        """
        return merge_data_relations(
            data=data,
            target_class=self._config.class_name,
            target_relation=self._config.target_relation,
            source_relations=self._config.source_relations,
        )


def check_model_assumptions_wrapper(
    model: SchemaPack, config: MergeRelationsConfig
) -> None:
    """Check the assumptions of the model.

    Raises:
        ModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    check_model_assumptions(
        model=model,
        class_name=config.class_name,
        target_relation=config.target_relation,
        source_relations=config.source_relations,
    )


def transform_model(model: SchemaPack, config: MergeRelationsConfig) -> SchemaPack:
    """Transform the data model.

    Raises:
        DataModelTransformationError:
            if the transformation fails.
    """
    return merge_model_relations(
        model=model,
        class_name=config.class_name,
        target_relation=config.target_relation,
        source_relations=config.source_relations,
        relation_spec=RelationSpecificationParams(
            mandatory=config.mandatory,
            multiple=config.multiple,
            description=config.description,
        ),
    )


MERGE_RELATIONS_TRANSFORMATION = TransformationDefinition[MergeRelationsConfig](
    config_cls=MergeRelationsConfig,
    check_model_assumptions=check_model_assumptions_wrapper,
    transform_model=transform_model,
    data_transformer_factory=MergeRelationsTransformer,
)


@register_transformation("merge_relations")
def get_merge_relations_transformation() -> TransformationDefinition[
    MergeRelationsConfig
]:
    """Get the merge relations transformation."""
    return MERGE_RELATIONS_TRANSFORMATION



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Models for configuration used in the 'merge relations' transformation."

import ast

from pydantic import Field, field_validator
from pydantic_settings import BaseSettings
from schemapack.spec.schemapack import MandatoryRelationSpec, MultipleRelationSpec


class MergeRelationsConfig(BaseSettings):
    """A model describing a configuration for merging relations.

    The `multiple` and the `mandatory` specifications are obtained from configuration.
    They are not inferred by the transformation.
    """

    class_name: str = Field(
        default=..., description="The name of the class to merge relations for."
    )
    description: str | None = Field(
        default=None, description="Description of the relation."
    )
    source_relations: list[str] = Field(
        default=..., description="List of relation names to be merged."
    )
    target_relation: str = Field(
        default=..., description="The name of the relation to merge into."
    )
    mandatory: MandatoryRelationSpec = Field(
        ..., description=("The modality of the relation.")
    )
    multiple: MultipleRelationSpec = Field(
        ..., description=("The cardinality of the relation.")
    )

    @field_validator("source_relations", mode="before")
    @classmethod
    def convert_source_relations(cls, value):
        """Handles Jinja2 rendering that stringifies lists: converts string source_relations
        into a list before pydantic validation.
        """
        if isinstance(value, str):
            return ast.literal_eval(value)
        return value



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data transformation logic for the `transform content` transformation"""

from collections.abc import Mapping

import yaml
from jinja2 import StrictUndefined
from jinja2.sandbox import SandboxedEnvironment
from schemapack import denormalize, isolate_resource
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.custom_types import EmbeddingProfile
from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.transform.exceptions import EvitableTransformationError

# configure with StrictUndefined so invalid property/dict access produces errors
# instead of silently inserting a none value
env = SandboxedEnvironment(undefined=StrictUndefined)

# object is too broad for what _denormalization_workaround will accept, define
# something that's closer to the intention of JsonObjectCompatible
ResourceContent = list | str | tuple | Mapping


def transform_data_content(
    *,
    data: DataPack,
    schemapack: SchemaPack,
    class_name: str,
    content_template_yaml: str,
    embedding_profile: EmbeddingProfile,
) -> DataPack:
    """Denormalize data using the configured class as root and perform content data transformation."""
    mutable_data = data_to_dict(data)

    if class_name not in data.resources:
        raise EvitableTransformationError()

    for resource_id in data.resources[class_name]:
        # while isolating and thus rooting the datapack here, the way this is applied
        # to all resources will result in a datapack that IS NOT rooted
        rooted_datapack = isolate_resource(
            datapack=data,
            class_name=class_name,
            resource_id=resource_id,
            schemapack=schemapack,
        )
        denormalized_content = denormalize(
            datapack=rooted_datapack,
            schemapack=schemapack,
            embedding_profile=embedding_profile,
        )

        id_property_name = schemapack.classes[class_name].id.propertyName
        content_property_names = list(
            schemapack.classes[class_name].content["properties"].keys()
        )
        relation_property_names = list(schemapack.classes[class_name].relations.keys())

        # evaluate data template using jinja
        transformed_content = env.from_string(content_template_yaml).render(
            original=denormalized_content,
            id_property_name=id_property_name,
            content_property_names=content_property_names,
            relation_property_names=relation_property_names,
        )

        # replace resource content with the transformed version
        mutable_data["resources"][class_name][resource_id]["content"] = yaml.safe_load(
            transformed_content
        )

    return DataPack.model_validate(mutable_data)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defining exceptions."""

import schemapack.exceptions


class ModelAssumptionError(RuntimeError):
    """Raised when assumptions made by transformation step about a model are not met."""


class MultiplicityError(ModelAssumptionError):
    """Raised when a relation in the model does not conform to the required multiplicity
    constraints. It occurs when the actual cardinality of a relationship within a model
    fails to meet the expected multiplicity criteria of a transformation. E.g.,
    in 'count references' transformation, the target of a relation is required to
    contribute multiple instances(`target=True`) to the relation, and this error is raised
    if that condition is not satisfied.
    """


class ModelTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to the schemapack-based model.
    This exception should only be raised when the error could not have been caught
    earlier by model assumption checks (otherwise the AssumptionsInsufficiencyError
    should be raised instead).
    """


class DataTransformationError(RuntimeError):
    """Raised when a transformation failed when applied to data in datapack-format.
    This exception should only be raised when the error could not have been caught
    earlier by model assumption checks (otherwise the EvitableTransformationError
    should be raised instead).
    """


class EvitableTransformationError(RuntimeError):
    """Raised when an exception during the model or data transformation should have
    been caught earlier by model assumption or data validation checks.
    """

    def __init__(self):
        super().__init__(
            "This unexpected error appeared during transformation, however, it should"
            + " have been caught earlier during model assumption checks (and/or by data"
            + " validation against the assumption-checked model). Please make sure that"
            + " the model assumption checks guarantee the workability of the"
            + " corresponding transformation wrt the provided model (and/or data)."
        )


class PreTransformValidationError(RuntimeError):
    """Raised when the validation of input data fails against the input model at the
    beginning of a data transformation.
    """

    def __init__(self, *, validation_error: schemapack.exceptions.ValidationError):
        """Initialize with the schemapack ValidationError."""
        super().__init__(
            "Validation of input data failed against the input model:"
            + f"\n{validation_error}"
        )


class PostTransformValidationError(RuntimeError):
    """Raised when the validation of transformed data fails against the transformed
    model at the end of a data transformation step.
    """

    def __init__(self, *, validation_error: schemapack.exceptions.ValidationError):
        """Initialize with the schemapack ValidationError."""
        super().__init__(
            "Validation of transformed data failed against the transformed model:"
            + f"\n{validation_error}"
        )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions to transform Schemapack models and Datapack data
into JSON-serializable dictionaries, as well as thawing frozen structures.
"""

import json
from typing import Any

from schemapack import dumps_datapack, dumps_schemapack
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import (
    ClassDefinition,
    ClassRelation,
    SchemaPack,
)

from metldata.builtin_transformations.common.path.path_elements import (
    RelationPathElement,
    RelationPathElementType,
)


def model_to_dict(model: SchemaPack) -> dict[str, Any]:
    """Converts the provided SchemaPack model to a dictionary."""
    return json.loads(dumps_schemapack(model, yaml_format=False))


def data_to_dict(data: DataPack) -> dict[str, Any]:
    """Converts the provided DataPack data to a dictionary."""
    return json.loads(dumps_datapack(data, yaml_format=False))


def content_to_dict(class_def: ClassDefinition) -> dict[str, Any]:
    """Converts a content schema into a dictionary."""
    class_def_dict = json.loads(class_def.model_dump_json())
    return class_def_dict["content"]


def relations_to_dict(class_def: ClassDefinition) -> dict[str, Any]:
    """Converts a relation schema into a dictionary."""
    class_def_dict = json.loads(class_def.model_dump_json())
    return class_def_dict["relations"]


def get_relation(element: RelationPathElement, schema: SchemaPack) -> ClassRelation:
    """Get the relation object for a path element.

    Args:
        element: The path element to get the relation for.
        schema: The underlying schema.

    Returns:
        The relation object.
    """
    element_active = element.type_ == RelationPathElementType.ACTIVE
    class_name = element.source if element_active else element.target
    return schema.classes[class_name].relations[element.property]



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Model transformation logic for the 'infer relation' transformation"

from schemapack.spec.schemapack import (
    ClassDefinition,
    ClassRelation,
    MandatoryRelationSpec,
    MultipleRelationSpec,
    SchemaPack,
)

from metldata.builtin_transformations.common.path import RelationPath
from metldata.builtin_transformations.common.path.path_elements import (
    RelationPathElementType,
)
from metldata.builtin_transformations.common.utils import get_relation, model_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def infer_multiplicity_from_path(
    path: RelationPath, schema: SchemaPack
) -> MultipleRelationSpec:
    """Infer the multiplicity of an inferred relation based on the path.

    Args:
        path: The path to infer the multiplicity for.
        schema: The underlying schema.

    Returns:
        The inferred multiplicity.
    """
    origin = target = False
    # Traverse the path and check for multiplicity
    for element in path.elements:
        relation = get_relation(element, schema)
        # If any multiplicity is observed, toggle the origin / source flag depending on
        # the orientation of the relation.
        if element.type_ == RelationPathElementType.ACTIVE:
            if relation.multiple.origin:
                origin = True
            if relation.multiple.target:
                target = True
        else:
            if relation.multiple.origin:
                target = True
            if relation.multiple.target:
                origin = True
        if origin and target:
            break
    return MultipleRelationSpec(origin=origin, target=target)


def infer_mandatory_from_path(
    path: RelationPath, schema: SchemaPack
) -> MandatoryRelationSpec:
    """Infer the mandatory property of an inferred relation based on the path.

    Args:
        path: The path to infer the mandatory property for.
        schema: The underlying schema.

    Returns:
        The inferred mandatory property.
    """
    origin = target = True
    # Traverse the path and check for mandatory
    for element in path.elements:
        relation = get_relation(element, schema)
        # If either end is not mandatory, toggle the origin / source flag depending on
        # the orientation of the relation.
        if element.type_ == RelationPathElementType.ACTIVE:
            if not relation.mandatory.origin:
                origin = False
            if not relation.mandatory.target:
                target = False
        else:
            if not relation.mandatory.origin:
                target = False
            if not relation.mandatory.target:
                origin = False
        if not origin and not target:
            break
    return MandatoryRelationSpec(origin=origin, target=target)


def infer_model_relation(
    *,
    model: SchemaPack,
    relation_name: str,
    relation_path: RelationPath,
) -> SchemaPack:
    """Add inferred relations to a model.

    Args:
        model: The model based on SchemaPack to add the inferred relations to.
        relation_name: The name of the new inferred relation.
        relation_path: The path to the relation to infer.

    Returns:
        The model with the inferred relation added.
    """
    updated_class_def: dict[str, ClassDefinition] = {}

    # This is the class that is being modified
    source_class = relation_path.source

    # This is the class that is referred by the path
    target_class = relation_path.target

    class_def = model.classes.get(source_class)

    if class_def is None:
        raise EvitableTransformationError()

    mandatory = infer_mandatory_from_path(relation_path, model)
    multiple = infer_multiplicity_from_path(relation_path, model)
    new_relation = ClassRelation.model_validate(
        {
            "targetClass": target_class,
            "mandatory": mandatory,
            "multiple": multiple,
        }
    )
    updated_class_def[source_class] = ClassDefinition.model_validate(
        {
            "id": class_def.id,
            "content": class_def.content,
            "relations": {
                **class_def.relations,
                relation_name: new_relation,
            },
        }
    )

    model_dict = model_to_dict(model)
    model_dict["classes"].update(updated_class_def)
    return SchemaPack.model_validate(model_dict)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Registry builder for built-in transformations."""

from collections.abc import Callable
from typing import Any, TypeVar

from metldata.transform.base import TransformationDefinition

T = TypeVar("T", bound=TransformationDefinition[Any])

_TRANSFORMATION_REGISTRY: dict[str, TransformationDefinition[Any]] = {}


def register_transformation(name: str) -> Callable[[Callable[[], T]], Callable[[], T]]:
    """Decorator to register a transformation under a given name."""

    def decorator(factory: Callable[[], T]) -> Callable[[], T]:
        if name in _TRANSFORMATION_REGISTRY:
            raise ValueError(f"Transformation '{name}' already registered")

        _TRANSFORMATION_REGISTRY[name] = factory()
        return factory

    return decorator


def get_transformation_registry() -> dict[str, TransformationDefinition[Any]]:
    """Get the registry of built-in transformations."""
    return _TRANSFORMATION_REGISTRY.copy()



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Model transformation logic for the 'merge relations' transformation"

from pydantic import BaseModel, Field
from schemapack.spec.schemapack import (
    MandatoryRelationSpec,
    MultipleRelationSpec,
    SchemaPack,
)

from metldata.builtin_transformations.common.custom_types import MutableClassRelations
from metldata.builtin_transformations.common.utils import model_to_dict
from metldata.transform.exceptions import EvitableTransformationError


class RelationSpecificationParams(BaseModel):
    """Parameters for relation specifications in the merge relations transformation."""

    mandatory: MandatoryRelationSpec = Field(
        ..., description="The modality of the relation."
    )
    multiple: MultipleRelationSpec = Field(
        ..., description="The cardinality of the relation."
    )
    description: str | None = Field(
        default=None, description="Description of the relation."
    )


def merge_model_relations(
    *,
    model: SchemaPack,
    class_name: str,
    target_relation: str,
    source_relations: list[str],
    relation_spec: RelationSpecificationParams,
) -> SchemaPack:
    """Model transformation logic for the 'merge relations' transformation.
    Args:
        model: The model to be transformed.
        class_name: The name of the class to merge relations for.
        target_relation: The name of the relation to merge into.
        source_relations: List of relation names to be merged.
        relation_spec: Parameters for the relation specification.
    """
    mutable_model = model_to_dict(model)

    target_class = _get_target_class(
        model=model, class_name=class_name, source_relations=source_relations
    )

    try:
        mutable_class_relations = mutable_model["classes"][class_name]["relations"]
    except KeyError as exc:
        raise EvitableTransformationError from exc

    # add the new relation
    mutable_class_relations[target_relation] = {
        "targetClass": target_class,
        "mandatory": relation_spec.mandatory,
        "multiple": relation_spec.multiple,
        "description": relation_spec.description,
    }

    # delete relations that are merged
    _delete_relations_from_model(
        class_relations=mutable_class_relations, source_relations=source_relations
    )

    return SchemaPack.model_validate(mutable_model)


def _get_target_class(
    *, model: SchemaPack, class_name: str, source_relations: list[str]
) -> str:
    """Retrieve the target class name from the first relation given in 'source relations'."""
    class_definition = model.classes.get(class_name)

    if class_definition is None:
        raise EvitableTransformationError()

    class_relations = class_definition.relations.get(source_relations[0])

    if class_relations is None:
        raise EvitableTransformationError()

    return class_relations.targetClass


def _delete_relations_from_model(
    *, class_relations: MutableClassRelations, source_relations: list[str]
) -> None:
    """Delete the relations specified in 'source relations' from the class_relations."""
    for relation_name in source_relations:
        try:
            del class_relations[relation_name]
        except KeyError as exc:
            raise EvitableTransformationError from exc



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Contains the transformation definition."""

from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.null.config import NullConfig
from metldata.transform.base import (
    DataTransformer,
    SubmissionAnnotation,
    TransformationDefinition,
)


def null_model_assumptions(model: SchemaPack, config: NullConfig):
    """No assumptions made."""
    return


def null_transform_model(model: SchemaPack, config: NullConfig) -> SchemaPack:
    """The model is returned unchanged."""
    return model


class NullTransformer(DataTransformer[NullConfig, SubmissionAnnotation]):
    """A Null transformer that returns the input model and data unchanged. Useful e.g.
    for testing.
    """

    def transform(self, data: DataPack, annotation: SubmissionAnnotation) -> DataPack:
        """Transforms data.

        Args:
            data: The data as DataPack to be transformed.

        Raises:
            DataTransformationError:
                if the transformation fails.
        """
        return data


NULL_TRANSFORMATION = TransformationDefinition[NullConfig](
    config_cls=NullConfig,
    check_model_assumptions=lambda schemapack, config: None,
    transform_model=lambda schemapack, config: schemapack,
    data_transformer_factory=NullTransformer,
)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models for configuration used in the 'rename id property' transformation."""

from pydantic import Field
from pydantic_settings import BaseSettings


class RenameIdPropertyConfig(BaseSettings):
    """A model describing a configuration for renaming model's id property."""

    class_name: str = Field(
        default=...,
        description="The name of the class whose id property will be renamed.",
    )
    id_property_name: str = Field(
        default=..., description="The name of the new id property."
    )
    description: str | None = Field(
        default=None,
        description="A description of the id property name.",
    )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for handling Jinja2 templates in workflows."""

from collections.abc import Mapping
from typing import Any

from jinja2.sandbox import ImmutableSandboxedEnvironment

# Random string, effectively disabling Jinja2 blocks and comments
RANDOM_START_END_STRING = "Wk4CPM:"
env = ImmutableSandboxedEnvironment(
    block_start_string=RANDOM_START_END_STRING + "A",
    block_end_string=RANDOM_START_END_STRING + "B",
    variable_start_string="{{{",
    variable_end_string="}}}",
    comment_start_string=RANDOM_START_END_STRING + "C",
    comment_end_string=RANDOM_START_END_STRING + "D",
)


def render_single_item(data: Any, item: Any) -> Any:
    """Recursively renders a variable named 'item' in a Json-compatible data structure."""
    if isinstance(data, str):
        return env.from_string(data).render(item=item)
    elif isinstance(data, Mapping):
        return {key: render_single_item(value, item) for key, value in data.items()}
    elif isinstance(data, list | tuple):
        return [render_single_item(value, item) for value in data]
    else:
        return data


def render_items(data: dict[str, Any], items: list[Any]) -> list[dict[str, Any]]:
    """Applies a loop to the data by rendering the template for each item."""
    return [render_single_item(data, item) for item in items]


def apply_loop(data: dict[str, Any]) -> list[dict[str, Any]]:
    """Applies a loop to the data by rendering the template for each item in the 'loop'
    key if present.
    """
    if not data.get("loop", False):
        return [data]
    items = data.pop("loop")
    return render_items(data, items)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A transformation to replace resource ids in the data."""

from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.registry import register_transformation
from metldata.builtin_transformations.replace_resource_ids.assumptions import (
    check_model_assumptions,
)
from metldata.builtin_transformations.replace_resource_ids.config import (
    ReplaceResourceIdsConfig,
)
from metldata.builtin_transformations.replace_resource_ids.data_transform import (
    replace_data_resource_ids,
)
from metldata.builtin_transformations.replace_resource_ids.model_transform import (
    replace_model_resource_ids,
)
from metldata.transform.base import (
    DataTransformer,
    SubmissionAnnotation,
    TransformationDefinition,
)


class ReplaceResourceIdsTransformer(
    DataTransformer[ReplaceResourceIdsConfig, SubmissionAnnotation]
):
    """A transformer that replaces the resource IDs in the data with the ones coming
    from an annotation.
    """

    def transform(self, data: DataPack, annotation: SubmissionAnnotation) -> DataPack:
        """Transforms data.
        The expected format of the 'annotation' parameter is:
        {
            'accession_map': {
                <class_name>: {
                    <old_id>: <new_id>,
                    ...
                },
                ...
            }
        }

        Args:
            data: The data as DataPack to be transformed.

        Raises:
            DataTransformationError:
                if the transformation fails.
        """
        return replace_data_resource_ids(
            data=data, class_name=self._config.class_name, annotation=annotation
        )


def check_model_assumptions_wrapper(
    model: SchemaPack, config: ReplaceResourceIdsConfig
) -> None:
    """Check the assumptions of the model.

    Raises:
        ModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    check_model_assumptions(model=model, class_name=config.class_name)


def transform_model(model: SchemaPack, config: ReplaceResourceIdsConfig) -> SchemaPack:
    """Transform the data model.

    Raises:
        DataModelTransformationError:
            if the transformation fails.
    """
    return replace_model_resource_ids(model=model)


REPLACE_RESOURCE_IDS_TRANSFORMATION = TransformationDefinition[
    ReplaceResourceIdsConfig
](
    config_cls=ReplaceResourceIdsConfig,
    check_model_assumptions=check_model_assumptions_wrapper,
    transform_model=transform_model,
    data_transformer_factory=ReplaceResourceIdsTransformer,
)


@register_transformation("replace_resource_ids")
def get_replace_resource_ids_transformation() -> TransformationDefinition[
    ReplaceResourceIdsConfig
]:
    """Get the replace resource ids transformation."""
    return REPLACE_RESOURCE_IDS_TRANSFORMATION



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration for the `transform content` transformation."""

from typing import Annotated, Any

import yaml
from pydantic import BeforeValidator, Field
from pydantic_settings import BaseSettings

from metldata.builtin_transformations.common.custom_types import EmbeddingProfile


def convert_dict_compatible(value: Any) -> Any:
    """Convert json serialized representation to dict."""
    if isinstance(value, str) and value.startswith("{"):
        return yaml.safe_load(value)
    return value


class TransformContentConfig(BaseSettings):
    """A config for arbitrary modifications of content schemas and data."""

    class_name: str = Field(
        default=...,
        description="Name of the class containing the content to be transformed.",
    )
    content_schema: Annotated[
        dict[str, object], BeforeValidator(convert_dict_compatible)
    ] = Field(
        default=...,
        description="JSON Schema the transformed content needs to adhere to.",
    )
    embedding_profile: Annotated[
        EmbeddingProfile, BeforeValidator(convert_dict_compatible)
    ] = Field(
        default_factory=dict,
        description="Embedding profile for denormalization. All relations are embedded by default and nested relations need to be specified explicitly for exclusion.",
    )
    content_template_yaml: str = Field(
        default=...,
        description="Jinja template used to transform existing content data.",
    )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Assumptions for the infer relation transformation."

from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.assumptions.path_assumptions import (
    assert_path_classes_and_relations_exist,
    assert_relation_does_not_exist,
)
from metldata.builtin_transformations.common.path import RelationPath


def check_model_assumptions(
    *,
    model: SchemaPack,
    class_name: str,
    relation_name: str,
    relation_path: RelationPath,
) -> None:
    """Check model assumptions for the infer relation transformation."""
    assert_path_classes_and_relations_exist(model=model, path=relation_path)
    # check if newly inferred relation already exists and if so, raise an error
    assert_relation_does_not_exist(
        model=model,
        class_name=class_name,
        relation_name=relation_name,
    )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model transformation logic for the 'delete class' transformation"""

from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.utils import model_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def delete_model_class(*, model: SchemaPack, class_name: str) -> SchemaPack:
    """Delete class from the model."""
    mutable_model = model_to_dict(model)

    try:
        del mutable_model["classes"][class_name]
    except KeyError as exc:
        raise EvitableTransformationError() from exc

    _remove_relations_from_model(
        mutable_model=mutable_model,
        original_model=model,
        target_class=class_name,
    )

    return SchemaPack.model_validate(mutable_model)


def _remove_relations_from_model(
    *,
    mutable_model: dict,
    original_model: SchemaPack,
    target_class: str,
) -> None:
    """Remove relations associated to deleted class from the model.

    Args:
        mutable_model (dict): Dictionary representation of the model
        original_model (SchemaPack): Original model
        target_class (str): Class name to be deleted
    """
    for class_name, class_def in original_model.classes.items():
        if class_name == target_class:
            continue

        filtered_relations = {
            rel_name: rel_spec
            for rel_name, rel_spec in class_def.relations.items()
            if rel_spec.targetClass != target_class
        }
        mutable_model["classes"][class_name]["relations"] = filtered_relations



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Combines functions controlling Path dependent common assumptions."

from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.path import RelationPath
from metldata.builtin_transformations.common.path.path_elements import (
    RelationPathElementType,
)
from metldata.builtin_transformations.common.utils import get_relation
from metldata.transform.exceptions import ModelAssumptionError, MultiplicityError


def assert_path_classes_and_relations_exist(
    *, model: SchemaPack, path: RelationPath
) -> None:
    """Make sure that all classes and relations defined in the provided path exist in
    the provided model.

    Raises:
        ModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    for path_element in path.elements:
        check_class_exists(model=model, class_name=path_element.source)
        check_class_exists(model=model, class_name=path_element.target)

        if path_element.type_ == RelationPathElementType.ACTIVE:
            check_relation_exists(
                model=model,
                class_name=path_element.source,
                relation=path_element.property,
            )

        if path_element.type_ == RelationPathElementType.PASSIVE:
            check_relation_exists(
                model=model,
                class_name=path_element.target,
                relation=path_element.property,
            )


def check_class_exists(*, model: SchemaPack, class_name: str) -> None:
    """Check if a class exists in the model and raise an error if not"""
    if class_name not in model.classes:
        raise ModelAssumptionError(f"Class {class_name} not found in model.")


def check_class_does_not_exist(*, model: SchemaPack, class_name: str) -> None:
    """Ensure a class does not exist in the model and raise an error else"""
    if class_name in model.classes:
        raise ModelAssumptionError(f"Unexcpected class {class_name} found in model.")


def check_relation_exists(*, model: SchemaPack, class_name: str, relation: str) -> None:
    """Check if a relation exists in a class and raise an error if not"""
    if relation not in model.classes[class_name].relations:
        raise ModelAssumptionError(
            f"Relation property {relation} not found in class {class_name}."
        )


def assert_relation_does_not_exist(
    *, model: SchemaPack, class_name: str, relation_name: str
) -> None:
    """Ensure a relation does not exist in a class and raise an error else."""
    if relation_name in model.classes[class_name].relations:
        raise ModelAssumptionError(
            f"Unexpected relation property {relation_name} found in class {class_name}."
        )


def assert_relation_target_multiplicity(
    *, model: SchemaPack, path: RelationPath
) -> None:
    """Make sure the target of the relation contributes multiple instances to the relation."""
    for element in path.elements:
        relation = get_relation(element, model)
        if element.type_ == RelationPathElementType.ACTIVE and relation.multiple.target:
            return
        if (
            element.type_ == RelationPathElementType.PASSIVE
            and relation.multiple.origin
        ):
            return

    raise MultiplicityError(
        f"Along the path {path} there is no target that contributes"
        + " multiple instances to the relation"
    )


def assert_no_relation_target_multiplicity(
    *, model: SchemaPack, path: RelationPath
) -> None:
    """Make sure relation targets contribute no more than one instance to the relation."""
    for element in path.elements:
        relation = get_relation(element, model)
        if element.type_ == RelationPathElementType.ACTIVE and relation.multiple.target:
            raise MultiplicityError(
                "Encountered target contributing multiple instances to a relation"
                + f" along path {path} while no multiplicity was expected."
            )
        if (
            element.type_ == RelationPathElementType.PASSIVE
            and relation.multiple.origin
        ):
            raise MultiplicityError(
                "Encountered target contributing multiple instances to a relation"
                + f" along path {path} while no multiplicity was expected."
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic for executing workflows."""

from typing import Any

from pydantic import BaseModel, ConfigDict
from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.transform.handling import TransformationHandler
from metldata.workflow.base import Workflow, WorkflowStep
from metldata.workflow.exceptions import UnknownTransformationError


class WorkflowResult(BaseModel):
    """Model and data after workflow execution."""

    model_config = ConfigDict(frozen=True)

    model: SchemaPack
    data: DataPack


class WorkflowStepHandler:
    """Handles the execution of a single workflow step by linking it to the appropriate
    transformation.

    Attributes:
    workflow_step (WorkflowStep): The workflow step to be executed.
    input_model (SchemaPack): The input model to be used for the transformation.
    """

    def __init__(self, workflow_step: WorkflowStep, input_model: SchemaPack):
        self.workflow_step = workflow_step
        self.input_model = input_model

    def execute(
        self,
        transformation_registry: dict[str, Any],
        *,
        validate_input: bool = False,
        validate_output: bool = False,
    ) -> TransformationHandler:
        """Executes the workflow step by retrieving the corresponding transformation
        from the registry and initializing a TransformationHandler.

        Raises a ValueError if the transformation is not found in the registry.
        """
        step_name = self.workflow_step.name
        step_args = self.workflow_step.args

        if step_name not in transformation_registry:
            raise UnknownTransformationError(
                f"Invalid transformation name. {step_name} does not exist in the "
                "transformation registry."
            )

        transformation_definition = transformation_registry[step_name]
        return TransformationHandler(
            transformation_definition=transformation_definition,
            transformation_config=transformation_definition.config_cls(**step_args),
            input_model=self.input_model,
            validate_input=validate_input,
            validate_output=validate_output,
        )


class WorkflowHandler[SubmissionAnnotation]:
    """Handles the execution of a workflow, applying a sequence of transformations
    to a datapack and associated schemapack.

    Attributes:
        workflow (Workflow): The workflow to be executed.
        transformation_registry (dict[str, Any]): A mapping from transformation names
            to their definitions.
        input_model (SchemaPack): The initial schema model that will be transformed by
            the workflow.

    """

    def __init__(
        self,
        workflow: Workflow,
        transformation_registry: dict[str, Any],
        input_model: SchemaPack,
    ):
        self.workflow = workflow
        self.transformation_registry = transformation_registry
        self.input_model = input_model

    def run(self, data: DataPack, annotation: SubmissionAnnotation) -> WorkflowResult:
        """Executes the workflow, applying each transformation in sequence to the model
        and data, and returns the final model and data as a WorkflowResult.
        """
        model = self.input_model

        last_step = len(self.workflow.operations) - 1
        for idx, step in enumerate(self.workflow.operations):
            step_handler = WorkflowStepHandler(workflow_step=step, input_model=model)

            # Collect and pass as kwargs.
            # More concise and works correctly if there's only one step in the workflow
            validation_args = {}
            if idx == 0:
                validation_args["validate_input"] = True
            if idx == last_step:
                validation_args["validate_output"] = True

            transformation_handler = step_handler.execute(
                self.transformation_registry, **validation_args
            )

            model = transformation_handler.transformed_model
            data = transformation_handler.transform_data(data, annotation)

        return WorkflowResult(model=model, data=data)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Assumptions for the merge relations transformation."

from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.assumptions.path_assumptions import (
    assert_relation_does_not_exist,
    check_class_exists,
    check_relation_exists,
)
from metldata.transform.exceptions import ModelAssumptionError


def check_model_assumptions(
    model: SchemaPack,
    class_name: str,
    target_relation: str,
    source_relations: list[str],
) -> None:
    """Check model assumptions for the merge relations transformation.
    Args:
        model: The model to be checked for assumptions.
        class_name: The name of the class to merge relations for.
        target_relation: The name of the relation to merge into.
        source_relations: List of relation names to be merged.
    """
    assert_class_and_relations_exists(
        model=model, class_name=class_name, source_relations=source_relations
    )
    # check if new relation already exists
    assert_relation_does_not_exist(
        model=model,
        class_name=class_name,
        relation_name=target_relation,
    )
    assert_same_target_class_across_relations(
        model=model, class_name=class_name, source_relations=source_relations
    )


def assert_class_and_relations_exists(
    *, model: SchemaPack, class_name: str, source_relations: list[str]
) -> None:
    """Make sure that the class and relations defined in the configuration exist in the
    provided model.

    Raises:
        ModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    check_class_exists(model=model, class_name=class_name)
    for relation_name in source_relations:
        check_relation_exists(
            model=model,
            class_name=class_name,
            relation=relation_name,
        )


def assert_same_target_class_across_relations(
    *,
    model: SchemaPack,
    class_name: str,
    source_relations: list[str],
):
    """Ensure that all source relations have the same target class."""
    class_relations = model.classes[class_name].relations
    target_classes = {
        class_relations[relation_name].targetClass for relation_name in source_relations
    }
    if len(target_classes) > 1:
        raise ModelAssumptionError(
            "All source relations must have the same target class, "
            f"but found: {target_classes}"
        )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic to build a workflow from a template."""

from metldata.workflow.base import (
    Workflow,
    WorkflowStep,
    WorkflowStepPrecursor,
    WorkflowTemplate,
)
from metldata.workflow.template_utils import apply_loop


class WorkflowBuilder:
    """A builder class for constructing Workflow objects from WorkflowTemplate instances.

    Attributes:
        template (WorkflowTemplate): The workflow template used as the basis for building the workflow.
    """

    def __init__(self, template: WorkflowTemplate):
        self.template = template

    def build(self) -> Workflow:
        """Build and return a Workflow object based on the template, with all loops expanded."""
        precursors = self.generate_step_precursors()
        operations = self.expand_loops(precursors)
        return Workflow(
            operations=operations,
        )

    def generate_step_precursors(self) -> list[WorkflowStepPrecursor]:
        """Generate workflow step precursors from the template's operations."""
        return [
            WorkflowStepPrecursor.model_validate(operation)
            for operation in self.template.operations
        ]

    def expand_loops(
        self, precursors: list[WorkflowStepPrecursor]
    ) -> list[WorkflowStep]:
        """Expand all loops in a list of workflow step precursors."""
        precursor_dicts = [
            precursor.model_dump(mode="json") for precursor in precursors
        ]
        step_dicts = [
            step
            for precursor_dict in precursor_dicts
            for step in apply_loop(precursor_dict)
        ]
        return [WorkflowStep.model_validate(step_dict) for step_dict in step_dicts]



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Built-in transformations"""

from metldata.builtin_transformations.delete_class import (
    DELETE_CLASS_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.duplicate_class import (
    DUPLICATE_CLASS_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.infer_relation import (
    INFER_RELATION_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.merge_relations import (
    MERGE_RELATIONS_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.rename_id_property import (
    RENAME_ID_PROPERTY_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.replace_resource_ids import (
    REPLACE_RESOURCE_IDS_TRANSFORMATION,  # noqa: F401
)
from metldata.builtin_transformations.transform_content import (
    TRANSFORM_CONTENT_TRANSFORMATION,  # noqa: F401
)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Assumptions for the `rename id property` transformation"

from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.common.assumptions.path_assumptions import (
    check_class_exists,
)
from metldata.transform.exceptions import ModelAssumptionError


def check_model_assumptions(
    model: SchemaPack, class_name: str, new_id_property: str
) -> None:
    """Check model assumptions for the `rename id property` transformation."""
    check_class_exists(model=model, class_name=class_name)
    check_id_is_different_from_existing(
        model=model,
        class_name=class_name,
        new_id_property=new_id_property,
    )


def check_id_is_different_from_existing(
    model: SchemaPack, class_name: str, new_id_property: str
) -> None:
    """Check that the current id is not the same as the new one."""
    existing_id_property = model.classes[class_name].id.propertyName
    if existing_id_property == new_id_property:
        raise ModelAssumptionError(
            f"The id property '{new_id_property}' already exists in the model."
        )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Logic for transforming data."

from typing import NamedTuple

from schemapack.spec.datapack import DataPack, Resource

from metldata.builtin_transformations.common.custom_types import ResourceId
from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.transform.exceptions import EvitableTransformationError


class Target(NamedTuple):
    """Model defining target elements."""

    target_resources: list[set[ResourceId] | ResourceId]
    target_class: str


def merge_data_relations(
    *,
    data: DataPack,
    target_class: str,
    target_relation: str,
    source_relations: list[str],
) -> DataPack:
    """Merge relations in the data according to the transformation configuration.
    Args:
        data: The data to be transformed.
        target_class: The name of the class to merge relations for.
        target_relation: The name of the relation to merge into.
        source_relations: List of relation names to be merged.
    """
    modified_data = data_to_dict(data)

    target_class_resources = data.resources.get(target_class)
    if target_class_resources is None:
        raise EvitableTransformationError()

    modified_resources = modified_data["resources"][target_class]

    for resource_id, resource in target_class_resources.items():
        targets = get_all_targets(resource, source_relations)

        modified_resources[resource_id]["relations"][target_relation] = {
            "targetClass": targets.target_class,
            "targetResources": set().union(*targets.target_resources),
        }

        # Remove source relations
        for relation_name in source_relations:
            del modified_resources[resource_id]["relations"][relation_name]

    return DataPack.model_validate(modified_data)


def get_all_targets(resource: Resource, source_relations: list[str]) -> Target:
    """Get target resources and the target class of the merged relations."""
    try:
        all_targets = []
        for relation_name in source_relations:
            target_resources = resource.relations[relation_name].targetResources
            if target_resources is not None:
                all_targets.append(target_resources)

        # must be same across the merging relations
        relation_target_class = resource.relations[source_relations[0]].targetClass

    except KeyError as exc:
        raise EvitableTransformationError() from exc

    return Target(target_resources=all_targets, target_class=relation_target_class)



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data transformation logic for the 'delete class' transformation"""

from schemapack.spec.datapack import DataPack

from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.transform.exceptions import EvitableTransformationError


def delete_data_class(*, data: DataPack, class_name: str) -> DataPack:
    """Delete class from the provided data."""
    modified_data = data_to_dict(data)

    try:
        del modified_data["resources"][class_name]
    except KeyError as exc:
        raise EvitableTransformationError() from exc

    _remove_relations_from_data(
        modified_data=modified_data,
        original_data=data,
        target_class=class_name,
    )
    return DataPack.model_validate(modified_data)


def _remove_relations_from_data(
    *, modified_data: dict, original_data: DataPack, target_class: str
) -> None:
    """Remove relations associated to deleted class from the data.

    Args:
        modified_data (dict): Dictionary representation of the data
        original_data (DataPack): Original data
        target_class (str): Name of the class to be deleted
    """
    for class_name, class_resources in original_data.resources.items():
        if class_name == target_class:
            continue

        for resource_id, resource in class_resources.items():
            filtered_relations = {
                name: specification
                for name, specification in resource.relations.items()
                if specification.targetClass != target_class
            }
            modified_data["resources"][class_name][resource_id]["relations"] = (
                filtered_relations
            )



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A collection of custom types used for builtin transformations."""

from collections.abc import Callable, Mapping
from typing import Annotated, Any, TypeAlias

from pydantic import Field

from metldata.builtin_transformations.common.path import RelationPath

_NonEmptyStr: TypeAlias = Annotated[str, Field(..., min_length=1)]
ResourceId: TypeAlias = _NonEmptyStr
MutableDatapack: TypeAlias = dict[str, Any]
MutableResource: TypeAlias = dict[str, dict]
MutableClassResources: TypeAlias = dict[ResourceId, MutableResource]
MutableResourceContent: TypeAlias = dict[str, Any]
ResolveRelations: TypeAlias = Callable[
    [ResourceId, RelationPath], frozenset[ResourceId]
]
MutableClassRelations: TypeAlias = dict[str, dict]
# needs to be a TypeAliasType so Pydantic can deal with the recursive definition
type EmbeddingProfile = Mapping[str, "bool | EmbeddingProfile"] | None
NewResourceId: TypeAlias = _NonEmptyStr
AccessionMap: TypeAlias = dict[ResourceId, NewResourceId]



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"Logic for transforming data."

from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import ClassDefinition, ClassRelation, SchemaPack

from metldata.builtin_transformations.common.custom_types import (
    MutableClassResources,
    MutableDatapack,
    ResourceId,
)
from metldata.builtin_transformations.common.path import RelationPath
from metldata.builtin_transformations.common.utils import data_to_dict
from metldata.builtin_transformations.infer_relation.resolve_path import (
    resolve_path,
)
from metldata.transform.exceptions import EvitableTransformationError


def infer_data_relation(
    *,
    data: DataPack,
    model: SchemaPack,
    class_name: str,
    relation_name: str,
    relation_path: RelationPath,
) -> DataPack:
    """Adds inferred relations to the data as given in the configuration.

    Args:
        data: The data to be transformed.
        model: The model of the data.
        class_name: The name of the class to which the relation will be added.
        relation_name: The name of the relation to be added.
        relation_path: The path indicating the route by which the two classes are related to each other.
    """
    modified_data = data_to_dict(data)

    referenced_class = relation_path.target

    # the target prefix refers to resources that will be modified by the transformation
    target_resources = get_class_resources(data=modified_data, class_name=class_name)

    for resource_id, resource in target_resources.items():
        relation_target_ids = resolve_path(
            data=data, source_resource_id=resource_id, path=relation_path
        )
        target_ids = conditionally_unpack_target_ids(
            model=model,
            class_name=class_name,
            relation_name=relation_name,
            relation_target_ids=relation_target_ids,
        )
        # add a new relation to that resource
        resource.setdefault("relations", {})[relation_name] = {
            "targetClass": referenced_class,
            "targetResources": target_ids,
        }
    return DataPack.model_validate(modified_data)


def get_class_resources(
    *, data: MutableDatapack, class_name: str
) -> MutableClassResources:
    """Extract the resources of a given class from the dictionary."""
    resources = data["resources"].get(class_name)
    if not resources:
        raise EvitableTransformationError()
    return resources


def conditionally_unpack_target_ids(
    *,
    model: SchemaPack,
    class_name: str,
    relation_name: str,
    relation_target_ids: frozenset[ResourceId],
) -> frozenset[ResourceId] | ResourceId | None:
    """Unpack resource ID from set if a string is expected by the model multiplicity
    specification.

    This function does not handle the case where `relation_target_ids` is expected
    to contain a single item, but `resolve_path` returns multiple values. Such cases
    will trigger a validation error in post-transformation validation.
    """
    class_definition = get_class_definition(model, class_name)

    relation_spec = get_relation_specification(class_definition, relation_name)

    # If relation expects single target and relation_target_ids has 0 or 1 item,
    # return single item or None
    expects_single_target = not relation_spec.multiple.target
    has_at_most_one_target = len(relation_target_ids) < 2
    if expects_single_target and has_at_most_one_target:
        return next(iter(relation_target_ids), None)

    return relation_target_ids


def get_class_definition(model: SchemaPack, class_name: str) -> ClassDefinition:
    """Get the class definition from the model by its name."""
    class_definition = model.classes.get(class_name)
    if class_definition is None:
        raise EvitableTransformationError()
    return class_definition


def get_relation_specification(
    class_def: ClassDefinition, relation_name: str
) -> ClassRelation:
    """Get the relation specification from the class definition by its name."""
    relation_multiplicity_specification = class_def.relations.get(relation_name)
    if relation_multiplicity_specification is None:
        raise EvitableTransformationError()
    return relation_multiplicity_specification



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A transformation to modify the content schema and data of a given class."""

from schemapack.spec.datapack import DataPack
from schemapack.spec.schemapack import SchemaPack

from metldata.builtin_transformations.registry import register_transformation
from metldata.builtin_transformations.transform_content.assumptions import (
    check_model_assumptions,
)
from metldata.builtin_transformations.transform_content.config import (
    TransformContentConfig,
)
from metldata.builtin_transformations.transform_content.data_transform import (
    transform_data_content,
)
from metldata.builtin_transformations.transform_content.model_transform import (
    transform_model_content,
)
from metldata.transform.base import (
    DataTransformer,
    SubmissionAnnotation,
    TransformationDefinition,
)


class TransformContentTransformer(
    DataTransformer[TransformContentConfig, SubmissionAnnotation]
):
    """A transformer that modifies the content schema and the data."""

    def transform(self, data: DataPack, annotation: SubmissionAnnotation) -> DataPack:
        """Transforms data.

        Args:
            data: The data as DataPack to be transformed.

        Raises:
            DataTransformationError:
                if the transformation fails.
        """
        return transform_data_content(
            data=data,
            schemapack=self._input_model,
            class_name=self._config.class_name,
            content_template_yaml=self._config.content_template_yaml,
            embedding_profile=self._config.embedding_profile,
        )


def check_model_assumptions_wrapper(
    model: SchemaPack, config: TransformContentConfig
) -> None:
    """Check the assumptions of the model.

    Raises:
        ModelAssumptionError:
            if the model does not fulfill the assumptions.
    """
    check_model_assumptions(model=model, class_name=config.class_name)


def transform_model(model: SchemaPack, config: TransformContentConfig) -> SchemaPack:
    """Transform the data model.

    Raises:
        DataModelTransformationError:
            if the transformation fails.
    """
    return transform_model_content(
        model=model, class_name=config.class_name, content_schema=config.content_schema
    )


TRANSFORM_CONTENT_TRANSFORMATION = TransformationDefinition[TransformContentConfig](
    config_cls=TransformContentConfig,
    check_model_assumptions=check_model_assumptions_wrapper,
    transform_model=transform_model,
    data_transformer_factory=TransformContentTransformer,
)


@register_transformation("transform_content")
def get_transform_content_transformation() -> TransformationDefinition[
    TransformContentConfig
]:
    """Get the transform content transformation."""
    return TRANSFORM_CONTENT_TRANSFORMATION



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Workflow validation logic."""

from pydantic import ValidationError

from metldata.transform.base import TransformationDefinition
from metldata.workflow.base import Workflow
from metldata.workflow.exceptions import WorkflowValidationError


def validate_workflow_against_registry(
    workflow: Workflow, transformation_registry: dict[str, TransformationDefinition]
) -> None:
    """Validate a workflow against a transformation registry.

    At this stage, the workflow has already been parsed into a ``Workflow``
    instance and is assumed to be structurally valid. This function performs
    semantic validation by ensuring that:

    1. Each referenced transformation exists in the registry.
    2. The provided operation arguments are compatible with the
    transformation's expected configuration class.

    Args:
        workflow: The workflow instance to validate.
        transformation_registry: Mapping of transformation names to their
            corresponding definitions.

    Raises:
        WorkflowValidationError: If a referenced transformation does not exist
            or if the provided configuration is incompatible with the
            transformation definition.
    """
    if not transformation_registry:
        raise WorkflowValidationError("Transformation registry cannot be empty.")

    for operation in workflow.operations:
        if operation.name not in transformation_registry:
            raise WorkflowValidationError(
                f"Unknown transformation '{operation.name}' referenced in workflow."
            )

        transformation_definition = transformation_registry[operation.name]
        provided_operation_config = operation.args
        expected_config_class = transformation_definition.config_cls

        # Validate by trying to create a valid config instance
        if not isinstance(provided_operation_config, expected_config_class):
            try:
                expected_config_class.model_validate(provided_operation_config)
            except ValidationError as error:
                raise WorkflowValidationError(
                    f"Invalid configuration for transformation '{operation.name}'. "
                    f"Provided arguments {provided_operation_config} are not "
                    f"compatible with expected config type "
                    f"{expected_config_class.__name__}."
                ) from error



# Copyright 2021 - 2026 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Models to describe workflow."""

from collections.abc import Mapping, Sequence
from typing import TypeVar

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings

TransformationConfig = TypeVar("TransformationConfig", bound=BaseSettings)


class WorkflowTemplate(BaseModel):
    """Base class for workflow templates."""

    operations: Sequence[Mapping[str, object]] = Field(
        default=...,
        description="List of operations to apply during the workflow in the given order.",
    )


class WorkflowStepBase(BaseModel):
    """Represents a single step within a workflow, defining a transformation to be applied."""

    name: str = Field(
        default=...,
        description="The name of the transformation that corresponds to an entry"
        + " in the transformation registry.",
    )
    description: str = Field(description="A description explaining the workflow step.")
    args: object = Field(
        default=...,
        description="The arguments or parameters required by the transformation."
        + " These are passed to the transformation definition and vary depending"
        + " on the transformation type.",
    )


class WorkflowStepPrecursor(WorkflowStepBase):
    """Represents a precursor model with loop specification by extending
    'WorkflowStepBase' with a loop attribute.
    """

    loop: Sequence[object] = Field(
        default_factory=list,
        description="Optional list specifying loop parameters for the workflow step.",
    )


class WorkflowStep[TransformationConfig](WorkflowStepBase):
    """Represents a workflow step model with typed transformation arguments."""

    args: TransformationConfig = Field(
        default=..., description="The configuration arguments for the workflow step."
    )


class Workflow(BaseModel):
    """Represents a workflow consisting of a sequence of transformation steps."""

    operations: Sequence[WorkflowStep] = Field(
        default=...,
        min_length=1,
        description="The ordered list of workflow steps to be applied to the input model.",
    )
